# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from View.ASubWindow import ASubMainWindow
from View.ListedSubWindow import ListedSubWindow

from PyQt5.QtGui import (
    QKeySequence,
)

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox, QVBoxLayout, QHeaderView, QTabWidget,
)

from View.SedimentLayers.UndoCommand import *
from View.SedimentLayers.Table import *

from View.SedimentLayers.Edit.Plot import Plot

from View.Plot.MplCanvas import MplCanvas
from View.SedimentLayers.translate import *

from View.SedimentLayers.Edit.Window import EditSedimentLayersWindow

_translate = QCoreApplication.translate

logger = logging.getLogger()

class SedimentLayersWindow(ASubMainWindow, ListedSubWindow):
    def __init__(self, title="SedimentLayersList", study=None, parent=None):
        self._study = study
        self._sediment_layers = self._study.river.sediment_layers

        self.setup_title(title)

        super(SedimentLayersWindow, self).__init__(
            name=self._title, ui="SedimentLayersList", parent=parent
        )

        self.setup_sc()
        self.setup_table()
        self.setup_graph()
        self.setup_connections()

        self.ui.setWindowTitle(self._title)

    def setup_title(self, title):
        self._title = (
            title + " - " + self._study.name
        )

    def setup_sc(self):
        self._undo_stack = QUndoStack()

        self.undo_sc = QShortcut(QKeySequence.Undo, self)
        self.redo_sc = QShortcut(QKeySequence.Redo, self)
        self.copy_sc = QShortcut(QKeySequence.Copy, self)
        self.paste_sc = QShortcut(QKeySequence.Paste, self)

    def setup_table(self):
        table = self.find(QTableView, f"tableView")
        self._table = TableModel(
            study = self._study,
            undo = self._undo_stack,
        )
        table.setModel(self._table)

        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        table.setAlternatingRowColors(True)

    def setup_graph(self):
        self.canvas = MplCanvas(width=5, height=4, dpi=100)
        self.canvas.setObjectName("canvas")
        self.plot_layout = self.find(QVBoxLayout, "verticalLayout")
        self.plot_layout.addWidget(self.canvas)

    def setup_connections(self):
        self.find(QAction, "action_add").triggered.connect(self.add)
        self.find(QAction, "action_del").triggered.connect(self.delete)
        self.find(QAction, "action_edit").triggered.connect(self.edit_sediment_layers)

        self.undo_sc.activated.connect(self.undo)
        self.redo_sc.activated.connect(self.redo)
        self.copy_sc.activated.connect(self.copy)
        self.paste_sc.activated.connect(self.paste)

        table = self.find(QTableView, f"tableView")
        table.selectionModel()\
             .selectionChanged\
             .connect(self._set_current_sl)

        self._table.dataChanged\
                   .connect(self._set_current_sl)

    def _set_current_sl(self):
        rows = self.index_selected_rows()

        if len(rows) == 0:
            return

        self.plot = Plot(
            canvas = self.canvas,
            data = self._sediment_layers.get(rows[0]),
            toolbar = None,
            display_current = False
        )
        self.plot.draw()


    def index_selected_rows(self):
        table = self.find(QTableView, f"tableView")
        return list(
            # Delete duplicate
            set(
                map(
                    lambda i: i.row(),
                    table.selectedIndexes()
                )
            )
        )

    def add(self):
        rows = self.index_selected_rows()
        if len(self._sediment_layers) == 0 or len(rows) == 0:
            self._table.add(0)
        else:
            self._table.add(rows[0])

    def delete(self):
        rows = self.index_selected_rows()
        if len(rows) == 0:
            return

        self._table.delete(rows)

    def copy(self):
        logger.info("TODO: copy")

    def paste(self):
        logger.info("TODO: paste")

    def undo(self):
        self._table.undo()

    def redo(self):
        self._table.redo()

    def edit_sediment_layers(self):
        rows = self.index_selected_rows()

        for row in rows:
            slw = EditSedimentLayersWindow(
                study = self._study,
                sl = self._sediment_layers.get(row),
                parent = self
            )
            slw.show()
